#!/usr/bin/env python

"""
Moin wiki format document tree nodes.

Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

class Container:

    "A container of document nodes."

    def __init__(self, nodes):
        self.nodes = nodes

    def append(self, node):
        self.nodes.append(node)

    append_text = append

    def empty(self):
        return not self.nodes

    def normalise(self):

        "Combine adjacent text nodes."

        nodes = self.nodes
        self.nodes = []
        text = None

        for node in nodes:

            # Open a text node or merge text into an open node.

            if isinstance(node, Text):
                if not text:
                    text = node
                else:
                    text.merge(node)

            # Close any open text node and append the current node.

            else:
                if text:
                    self.append(text)
                    text = None
                self.append(node)

        # Add any open text node.

        if text:
            self.append(text)

    def __str__(self):
        return self.prettyprint()

    def prettyprint(self, indent=""):
        pass

class Region(Container):

    "A region of the page."

    transparent_region_types = ["wiki"]

    def __init__(self, nodes, level=0, indent=0, type=None):
        Container.__init__(self, nodes)
        self.level = level
        self.indent = indent
        self.type = type

    def append(self, node):
        last = self.nodes and self.nodes[-1]
        if last and last.empty():
            self.nodes[-1] = node
        else:
            self.nodes.append(node)

    def append_text(self, s):
        if self.is_transparent():
            self.nodes[-1].append(s)
        else:
            self.append(s)

    def have_end(self, s):
        return self.level and s.startswith("}") and self.level == len(s)

    def is_transparent(self):
        return not self.level or self.type in self.transparent_region_types

    def __repr__(self):
        return "Region(%r, %r, %r, %r)" % (self.nodes, self.level, self.indent, self.type)

    def prettyprint(self, indent=""):
        l = ["%sRegion: level=%d indent=%d type=%s" % (indent, self.level, self.indent, self.type)]
        for node in self.nodes:
            l.append(node.prettyprint(indent + "  "))
        return "\n".join(l)

    def to_string(self, out):
        out.start_region(self.level, self.indent, self.type)
        for node in self.nodes:
            node.to_string(out)
        out.end_region(self.level, self.indent, self.type)

class Block(Container):

    "A block in the page."

    def __init__(self, nodes, final=True):
        Container.__init__(self, nodes)
        self.final = final

    def __repr__(self):
        return "Block(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sBlock: final=%s" % (indent, self.final)]
        for node in self.nodes:
            l.append(node.prettyprint(indent + "  "))
        return "\n".join(l)

    def to_string(self, out):
        out.start_block(self.final)
        for node in self.nodes:
            node.to_string(out)
        out.end_block(self.final)

class Heading(Container):

    "A heading."

    def __init__(self, nodes, level, start_extra="", start_pad="", end_pad="", end_extra=""):
        Container.__init__(self, nodes)
        self.level = level
        self.start_extra = start_extra
        self.start_pad = start_pad
        self.end_pad = end_pad
        self.end_extra = end_extra

    def __repr__(self):
        return "Heading(%r, %d, %r, %r, %r, %r)" % (
            self.nodes, self.level, self.start_extra, self.start_pad, self.end_pad, self.end_extra)

    def prettyprint(self, indent=""):
        l = ["%sHeading: level=%d start_extra=%r start_pad=%r end_pad=%r end_extra=%r" % (
                indent, self.level, self.start_extra, self.start_pad, self.end_pad, self.end_extra)]
        for node in self.nodes:
            l.append(node.prettyprint(indent + "  "))
        return "\n".join(l)

    def to_string(self, out):
        out.start_heading(self.level, self.start_extra, self.start_pad)
        for node in self.nodes:
            node.to_string(out)
        out.end_heading(self.level, self.end_pad, self.end_extra)

class ListItem(Container):

    "A list item."

    def __repr__(self):
        return "ListItem(%r)" % self.nodes

    def prettyprint(self, indent=""):
        l = ["%sListItem:" % indent]
        for node in self.nodes:
            l.append(node.prettyprint(indent + "  "))
        return "\n".join(l)

    def to_string(self, out):
        out.start_listitem()
        for node in self.nodes:
            node.to_string(out)
        out.end_listitem()



class Node:

    "A document node without children."

    def empty(self):
        return False

class Rule(Node):

    "A horizontal rule."

    def __init__(self, length):
        self.length = length

    def __repr__(self):
        return "Rule(%d)" % self.length

    def prettyprint(self, indent=""):
        return "%sRule: %d" % (indent, self.length)

    def to_string(self, out):
        out.rule(self.length)

class Text(Node):

    "A text node."

    def __init__(self, s):
        self.s = s

    def empty(self):
        return not self.s

    def merge(self, text):
        self.s += text.s

    def __repr__(self):
        return "Text(%r)" % self.s

    def prettyprint(self, indent=""):
        return "%sText: %r" % (indent, self.s)

    def to_string(self, out):
        out.text(self.s)

# vim: tabstop=4 expandtab shiftwidth=4
